/******************************************************************************
 * file: DoubleLinkedList.java
 * 
 * description: provides a class to create double linked lists and methods
 * to perform typical operations on these type of data set
 *****************************************************************************/


import java.util.Iterator;


/* class Index */
class Index {
  /* variables */
  private int predIndex;
  private int nextIndex;

  /* constructor */
  public Index() {
    this.predIndex = 0;
    this.nextIndex = 1;
  }

  /* returns the value of prev */
  public int prev() {
    return this.predIndex;
  }

  /* returns the value of next */
  public int next() {
    return this.nextIndex;
  }

  /* toggle */
  public void toggle() {
    this.predIndex = this.nextIndex;
    this.nextIndex = 1 - this.nextIndex;
  }
}


/* a node of a double linked list has a content and a successor */  
class Node_DLL {


  /* variables */
  private Object content;
  private Node_DLL linkedNode[];
  private Index index;


  /* constructors */
  public Node_DLL(Object o, Index index) {
    this.linkedNode = new Node_DLL[2];
    this.setIndex(index);
    this.setContent(o);
    this.setNextNode(null);
    this.setPreviousNode(null);
  }

  public Node_DLL(Object o, Node_DLL p, Node_DLL n, Index index) {
    this.linkedNode = new Node_DLL[2];
    this.setIndex(index);
    this.setContent(o);
    this.setNextNode(n);
    this.setPreviousNode(p);
  }

  private void setIndex(Index index){
    this.index = index;
  }

  /* returns the content of this node */
  public Object getContent() {
    return this.content;
  }


  /* sets the content of this node to o */
  public void setContent(Object o) {
    this.content = o;
  }


  /* returns the predecessor of this node */
  public Node_DLL getPreviousNode() {
    return this.linkedNode[this.index.prev()];
  }


  /* sets the predecessor of this node to p */
  public void setPreviousNode(Node_DLL p) {
    this.linkedNode[this.index.prev()] = p;
  }


  /* returns the successor of this node */
  public Node_DLL getNextNode() {
    return this.linkedNode[this.index.next()];
  }


  /* sets the successor of this node to n */
  public void setNextNode(Node_DLL n) {
    this.linkedNode[this.index.next()] = n;
  }


  /* returns the node's content as a string */
  public String toString() {
    return (String) this.content;
  }
}


/* a double linked list */
public class DoubleLinkedList {


  /* variables */
  protected Node_DLL head;
  protected Node_DLL tail;
  protected Index index;


  /* constructor */
  public DoubleLinkedList() {
    this.head = null;
    this.tail = null;
    this.index = new Index();
  }


  /* returns true if this list is empty, otherwise false */
  public boolean isEmpty() {
    return (this.head == null);
  }


  /* returns the reference to the node, that contains Object o, */
  /* if this list contains o, otherwise null */
  public Node_DLL search(Object o) {
    Node_DLL tmp = this.head;
    while (tmp != null) {
      if (tmp.getContent().equals(o)) {
        return tmp;
      }
      tmp = tmp.getNextNode();
    }
    return null;
  }


  /* inserts a Node that contains Object o at the head*/
  public void insertHead(Object o) {
    if (this.isEmpty())
      this.head = this.tail = new Node_DLL(o, null, null, this.index);
    else {
      this.head.setPreviousNode(
          new Node_DLL(o, null, this.head, this.index));
      this.head = this.head.getPreviousNode();
    }
  }

  /* inserts a Node that contains Object o at the tail*/
  public void insertTail(Object o) {
    reverse();
    insertHead(o);
    reverse();
  }


  /* inserts a Node that contains Object o at the given position*/
  public void insertAt(Object o, Node_DLL pos) {
    Node_DLL predecessor = pos.getPreviousNode();
    if (predecessor != null) {
      Node_DLL tmp = new Node_DLL(o, predecessor, pos, index);
      predecessor.setNextNode(tmp);
      pos.setPreviousNode(tmp);
    } else {
      insertHead(o);
    }
  }


  /* removes the first Node that contains Object o */
  public void remove(Object o) {
    Node_DLL tmp = this.head;
    while (tmp != null) {
      if (tmp.getContent().equals(o)) {
        Node_DLL predecessor = tmp.getPreviousNode();
        Node_DLL successor = tmp.getNextNode();
        if (predecessor == null) {
          this.head = successor;
          successor.setPreviousNode(null);
          break;
        } else {
          if (successor == null) {
            this.tail = predecessor;
            predecessor.setNextNode(null);
          } else {
            predecessor.setNextNode(successor);
            successor.setPreviousNode(predecessor);
          }
          break;
        }

      }
      tmp = tmp.getNextNode();
    }
  }



  /* reverses the list */
  public void reverse() {
    Node_DLL tmp = this.head;
    this.head = this.tail;
    this.tail = tmp;
    this.index.toggle();
  }


  /* needed in order to implement the interface Iterator */
  public Iterator iterator() {
    return new DoubleLinkedListIterator(this);
  }


  /* returns the list's content as a string object */
  public String toString() {
    String str = "Head -";
    if (isEmpty()) {
      str += "> <- Tail";
    } else {
      Iterator i = this.iterator();
      while (i.hasNext()) {
        str += "> " + i.next() + " <-";
      }
      str += " Tail";
    }
    return str;
  }
}


/* implements the interface Iterator */
class DoubleLinkedListIterator implements Iterator {

  /* variables */
  private DoubleLinkedList list;
  private Node_DLL enumPointer;

  /* constructor */
  public DoubleLinkedListIterator(DoubleLinkedList list) {
    this.list = list;
    this.enumPointer = list.head;
  }

  /* returns true, if the list has more elements, else false */
  public boolean hasNext() {
    return (enumPointer != null);
  }

  /* returns the next element */
  public Object next() {
    Node_DLL tmp = enumPointer;
    enumPointer  = tmp.getNextNode();
    return tmp.getContent();
  }

  public void remove() {
    //This method has to be implemented!
  }
}



























